__version__ = '1.0.0'

'''
version history:

0.1.0: first publishing, May 11, 2025
0.1.1: fix dependency errors
0.1.2: fixing module not find error
0.1.3: restructuring package
0.1.4 ~ 0.1.6: fixing moviepy import issue
0.1.7: changed name back to 'piegy'
0.1.8: updated installation in README
0.1.9: first round of full debugging

1.0.0: first version in PyPI
1.1.0: debugging. Updated a range of functions, in the following modules: figures, videos, test_var, model, figure_tools
1.1.1: minor debugging in model module.

'''